.686
.model flat,stdcall
option casemap:none

include .\stdlib.inc

.code

OPTION PROLOGUE:NONE
OPTION EPILOGUE:NONE

; AMD MANUAL
; 64-Bit Signed Remainder Computation
; _llrem divides two signed 64-bit numbers and returns the remainder.
;
; In:       [ESP+8]:[ESP+4] = dividend
;           [ESP+16]:[ESP+12] = divisor
;
; Out:      EDX:EAX = remainder of division
;
; Destroys: EAX, ECX, EDX, EFlags
Int64Rem proc ull1:QWORD,ull2:QWORD
	push ebx               ; Save EBX as per calling convention.
	push esi               ; Save ESI as per calling convention.
	push edi               ; Save EDI as per calling convention.
	mov  ecx, [esp+28]     ; divisor-hi
	mov  ebx, [esp+24]     ; divisor-lo
	mov  edx, [esp+20]     ; dividend-hi
	mov  eax, [esp+16]     ; dividend-lo
	mov  esi, edx          ; sign(remainder) == sign(dividend)
	sar  esi, 31           ; (remainder < 0) ? -1 : 0
	mov  edi, edx          ; dividend-hi
	sar  edi, 31           ; (dividend < 0) ? -1 : 0
	xor  eax, edi          ; If (dividend < 0),
	xor  edx, edi          ;  compute 1's complement of dividend.
	sub  eax, edi          ; If (dividend < 0),
	sbb  edx, edi          ;  compute 2's complement of dividend.
	mov  edi, ecx          ; divisor-hi
	sar  edi, 31           ; (divisor < 0) ? -1 : 0
	xor  ebx, edi          ; If (divisor < 0),
	xor  ecx, edi          ;  compute 1's complement of divisor.
	sub  ebx, edi          ; If (divisor < 0),
	sbb  ecx, edi          ;  compute 2's complement of divisor.
	jnz  sr_big_divisor    ; divisor > 2^32 - 1
	cmp  edx, ebx          ; Only one division needed (ECX = 0)?
	jae  sr_two_divs       ; No, need two divisions.
	div  ebx               ; EAX = quotient_lo
	mov  eax, edx          ; EAX = remainder_lo
	mov  edx, ecx          ; EDX = remainder_lo = 0
	xor  eax, esi          ; If (remainder < 0),
	xor  edx, esi          ;  compute 1's complement of result.
	sub  eax, esi          ; If (remainder < 0),
	sbb  edx, esi          ;  compute 2's complement of result.
	pop  edi               ; Restore EDI as per calling convention.
	pop  esi               ; Restore ESI as per calling convention.
	pop  ebx               ; Restore EBX as per calling convention.
	ret                    ; Done, return to caller.
sr_two_divs:
	mov ecx, eax        ; Save dividend_lo in ECX.
	mov eax, edx        ; Get dividend_hi.
	xor edx, edx        ; Zero-extend it into EDX:EAX.
	div ebx             ; EAX = quotient_hi, EDX = intermediate remainder
	mov eax, ecx        ; EAX = dividend_lo
	div  ebx            ; EAX = quotient_lo
	mov  eax, edx       ; remainder_lo
	xor  edx, edx       ; remainder_hi = 0
	jmp  sr_makesign    ; Make remainder signed.
sr_big_divisor:
	sub  esp, 16             ; Create three local variables.
	mov  [esp], eax          ; dividend_lo             
	mov  [esp+4], ebx        ; divisor_lo             
	mov  [esp+8], edx        ; dividend_hi
	mov  [esp+12], ecx       ; divisor_hi
	mov  edi, ecx            ; Save divisor_hi.
	shr  edx, 1              ; Shift both
	rcr  eax, 1              ;  divisor and
	ror  edi, 1              ;  and dividend
	rcr  ebx, 1              ;  right by 1 bit.
	bsr  ecx, ecx            ; ECX = number of remaining shifts
	shrd ebx, edi, cl        ; Scale down divisor and
	shrd eax, edx, cl        ;  dividend such that divisor is
	shr  edx, cl             ;  less than 2^32 (that is, fits in EBX).
	rol  edi, 1              ; Restore original divisor_hi.
	div  ebx                 ; Compute quotient.
	mov  ebx, [esp]          ; dividend_lo
	mov  ecx, eax            ; Save quotient.
	imul edi, eax            ; quotient * divisor high word (low only)
	mul  DWORD PTR [esp+4]   ; quotient * divisor low word
	add  edx, edi            ; EDX:EAX = quotient * divisor
	sub  ebx, eax            ; dividend_lo - (quot.*divisor)_lo
	mov  ecx, [esp+8]        ; dividend_hi
	sbb  ecx, edx            ; Subtract divisor * quot. from dividend.
	sbb  eax, eax            ; remainder < 0 ? 0xffffffff : 0
	mov  edx, [esp+12]       ; divisor_hi
	and  edx, eax            ; remainder < 0 ? divisor_hi : 0
	and  eax, [esp+4]        ; remainder < 0 ? divisor_lo : 0
	add  eax, ebx            ; remainder_lo
	add  edx, ecx            ; remainder_hi
	add  esp, 16             ; Remove local variables.
sr_makesign:
	xor eax, esi   ; If (remainder < 0),
	xor edx, esi   ;  compute 1's complement of result.
	sub eax, esi   ; If (remainder < 0),
	sbb edx, esi   ;  compute 2's complement of result.
	pop edi        ; Restore EDI as per calling convention.
	pop esi        ; Restore ESI as per calling convention.
	pop ebx        ; Restore EBX as per calling convention.
	ret 4*4        ; Done, return to caller.
Int64Rem endp   

OPTION PROLOGUE:PROLOGUEDEF 
OPTION EPILOGUE:EPILOGUEDEF

end
